/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.0, (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.erlang.org/EPL1_0.txt
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Erlang-4.7.3, December, 1998.
 * 
 * The Initial Developer of the Original Code is Ericsson Telecom
 * AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
 * Telecom AB. All Rights Reserved.
 * 
 * Contributor(s): ______________________________________.''
 */
/* Created:  1 Aug 1995 by tobbe@erix.ericsson.se
 *
 * Function: I/O functions for the IG package.
 *
 * RCS: $Id: igio.c,v 1.4 1996/04/11 08:36:58 tobbe Exp tobbe $
 */
#include "erl_interface.h"
#include "ig.h"

extern int ig_fun_in;
extern int ig_errno;
extern void ig_dispatch(void);

static char* packet_buf = NULL;
static int packet_buf_size = 0;

typedef (*ReadWriteFunc)(int fd, char* buf, int size);

static int erl_gen_read_packet(int header_size, ReadWriteFunc reader,
			       int fd, char** buf);
static int erl_gen_write_packet(int header_size, ReadWriteFunc writer,
				int fd, char *buf, int size);
static int read_fill(ReadWriteFunc reader, int fd, char* buf, int len);
static int write_fill(ReadWriteFunc writer, int fd, char* buf, int len);

extern int read(int, char*, int);
extern int write(int, char*, int);

/* For internal use within IG */
char *igbufp;
int igbuflen;

int
ig_read(int fd)
{
    return igbuflen = erl_gen_read_packet(2, &read, fd, &igbufp);
}

int
ig_write(int fd, char *bp, int len)
{
    return erl_gen_write_packet(2, &write, fd, bp, len);
}

/*
 * Read from Erlang and dispatch the function call.
 */
int ig_call(void)
{
  if (ig_read( ig_fun_in ) > 0) {
    ig_dispatch();
    return ig_errno;
  }
  else 
    return IG_READ_ERROR;
}

static int
erl_gen_write_packet(int header_size, ReadWriteFunc writer,
		     int fd, char *buf, int size)
{
    char header[4];		/* Buffer for header. */
    int status;			/* Status from read operation. */
    int packet_size;		/* Size of packet. */
    char* p;

    static int masks[] = { 0, 0xffffff00, 0xffff0000, 0xff000000, 0 };

    /*
     * Check that the packet fits with the given header size.
     */

    if (masks[header_size] & size) {
	return -1;
    }

    /*
     * Encode the packet header.
     */

    p = header;
    switch (header_size) {
    case 4:
	*p++ = size >> 24;
	*p++ = size >> 16;
    case 2:
	*p++ = size >> 8;
    case 1:
	*p++ = size & 255;
    }

    /*
     * Write header and packet.
     */

    if ((status = write_fill(writer, fd, header, header_size)) < 0) {
	return status;
    }
    return write_fill(writer, fd, buf, size);
}

static int
erl_gen_read_packet(int header_size, ReadWriteFunc reader, int fd, char** buf)
{
    unsigned char header[4];	/* Buffer for header. */
    int status;			/* Status from read operation. */
    int packet_size;		/* Size of packet. */
    int i;

    /*
     * Read the header and convert it to a number.
     */

    status = read_fill(reader, fd, (char *) header, header_size);
    if (status <= 0) {
	return status;
    }
    packet_size = 0;
    for (i = 0; i < header_size; i++) {
	packet_size = packet_size << 8 | header[i];
    }
    
    /*
     * If no buffer allocated or too small, allocate buffer for the packet.
     */

    if (packet_buf == NULL) {
	packet_buf = (char *) malloc(packet_size);
	packet_buf_size = packet_size;
    } else if (packet_buf_size < packet_size) {
	packet_buf = (char *) realloc(packet_buf, packet_size);
	packet_buf_size = packet_size;
    }
    *buf = packet_buf;
    return read_fill(reader, fd, packet_buf, packet_size);
}

static int
read_fill(ReadWriteFunc reader, int fd, char* buf, int len)
{
    int i;
    int got;

    for (got = 0; got < len; got +=i) {
	if ((i = reader(fd, buf+got, len-got)) <= 0) {
	    return i;
	}
    }
    return len;
}

static int
write_fill(ReadWriteFunc writer, int fd, char* buf, int len)
{
    int i;
    int written;

    for (written = 0; written < len; written +=i) {
	if ((i = writer(fd, buf+written, len-written)) <= 0) {
	    return -1;
	}
    }
    return 0;
}
