/* ``The contents of this file are subject to the Erlang Public License,
 * Version 1.0, (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.erlang.org/EPL1_0.txt
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Erlang-4.7.3, December, 1998.
 * 
 * The Initial Developer of the Original Code is Ericsson Telecom
 * AB. Portions created by Ericsson are Copyright (C), 1998, Ericsson
 * Telecom AB. All Rights Reserved.
 * 
 * Contributor(s): ______________________________________.''
 */
/* Copyright (C) 1995, Ellemtel Telecommunications Systems Laboratories
 *
 * Created:  22 Jun 1995 by tobbe@erix.ericsson.se
 *
 * Function: Setup socket connections for the IG interface.
 *           
 * Exports:  igsock_connect/2 , igsock_listen/1, igsock_accept/1
 *
 * RCS: $Id: igsock.c,v 1.4 1996/05/29 19:20:19 tobbe Exp $ 
 */
#ifdef __WIN32__
#include <winsock2.h>
#else /* not __WIN32__ */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/time.h>
#endif

#include <string.h>
#include <fcntl.h>
#include "ig.h"

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff   /* should be in <netinet/in.h> */
#endif

#define IGSOCK_ERROR -1

#ifdef SUNOS4
extern unsigned long inet_addr();
extern void *memset();
extern int socket();
extern int connect();
extern int setsockopt();
extern int bind();
extern int listen();
extern int accept();
extern int fprintf();
extern void perror();
extern int select();
extern int close();
#endif

#ifdef SOLARIS
unsigned long inet_addr();
extern int close();
#endif

extern void igsetup_fun(int,int);
extern void igsetup_cback(int,int);

/* IGSOCK_CONNECT:
 *   Try to connect ourselves to an Erlang system, and initialize IG.
 */
int igsock_connect(char *host, /* Name or dotted-decimal addr of other system (Erlang) */
		   short port) /* Port number */
{
  static char old_host[255];
  static short old_port;
  static struct sockaddr_in remote_addr;
  struct hostent *hp;
  unsigned long inaddr;
  int sock_fd,rc;

  /* Check if we have to fill in the address information ! */
  if (strcmp(old_host,host) && (old_port != port)) {
    /* Yes, fill in the address to the other system */
    memset((void *) &remote_addr , 0 , (size_t) sizeof(remote_addr));
    remote_addr.sin_family = AF_INET;              
    remote_addr.sin_port   = htons( port ); 

    if ((inaddr = inet_addr( host )) != INADDR_NONE) 
      /* It's dotted decimal */
      memcpy((void *) &remote_addr.sin_addr.s_addr,  
	     (const void *) &inaddr ,
	     (size_t) sizeof(inaddr));
    else {
      if ((hp = gethostbyname( host )) == (struct hostent*) 0) 
	return IGSOCK_ERROR;
      else 
	/* found the address by name */
	memcpy((void *) &remote_addr.sin_addr.s_addr,  
	       (const void *) hp->h_addr ,
	       (size_t) hp->h_length);
    }
  }

  /* Create the remote socket */ 
  if ((sock_fd = socket( AF_INET , SOCK_STREAM , 0 )) < 0) 
    return IGSOCK_ERROR;
  
  /* Try to connect to the other system */
  if ((rc = connect(sock_fd, (struct sockaddr *) &remote_addr, sizeof(remote_addr))) < 0) { 
    closesocket( sock_fd ); 
    return IGSOCK_ERROR;
  }
  else {
    igsetup_cback( sock_fd , sock_fd );
    return sock_fd;
  }

} /* igsock_connect */

/* IGSOCK_LISTEN:
 *   Use this function to set up a Listen socket. When the socket
 *   triggers use the function igsock_accept/1 to setup the connection.
 */
int igsock_listen(short port) /* Port number */
{
  int listen_fd,one=1;
  struct sockaddr_in addr;

  if ( (listen_fd = socket( AF_INET , SOCK_STREAM , 0 )) < 0) 
    return IGSOCK_ERROR;

  memset((void *) &addr , 0 , (size_t) sizeof(addr));
  addr.sin_family = AF_INET;              
  addr.sin_port   = htons( port );   
  addr.sin_addr.s_addr = htonl( INADDR_ANY ); 

  if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, (char *) &one, sizeof(one)) < 0)
    return IGSOCK_ERROR;

  if (bind(listen_fd, (struct sockaddr *) &addr, sizeof( addr )) < 0)
    return IGSOCK_ERROR;
  
  listen(listen_fd, 5);
  return listen_fd;

} /* igsock_listen */


/* IGSOCK_ACCEPT:
 *   Set up the actual connection, and initialize IG.
 */
int igsock_accept(int listen_fd)
{
  int fd,cli_len;
  struct sockaddr_in cli_addr;

  cli_len = sizeof(cli_addr);
  if ((fd = accept( listen_fd , (struct sockaddr *) &cli_addr, &cli_len )) < 0)
    return IGSOCK_ERROR;
  else {
    igsetup_fun( fd , fd );
    return fd;
  }

  return IGSOCK_ERROR;

} /* igsock_accept */
